; ============================================================
; Watch battery voltage and charge current
;
; ATtiny13 Fuse High Byte: 0xff --> 0xf9 (BODLEVEL = 4.3V)
;
; Calibration:
; o apply 8.00 volt to charger input
; o short pins 6 and 7 for a short moment
; o the green LED should turn on for 2 seconds
; o apply 12.00 volt to charger input
; o short pins 6 and 7 for a short moment
; o the red LED should turn on for 2 seconds
;
; Copyright (C) 2005, 2006 Dick Streefland
;
; This is free software, licensed under the terms of the GNU General
; Public License as published by the Free Software Foundation.
; ============================================================

.section .text

.equ	CLOCK,		1200000	; internal clock frequency (Hz)
.equ	TIMER,		100	; timer interrupt rate (Hz)
.equ	ADC_PRESCALER,	4	; 1200000 / 16 = 75000

.equ	ADCSRB,	0x03		; ADC Control and Status Register B
.equ	ADCL,	0x04		; ADC Data Register Low
.equ	ADCH,	0x05		; ADC Data Register High
.equ	ADCSRA,	0x06		; ADC Control and Status Register A
.equ	ADMUX,	0x07		; ADC Multiplexer Selection Register
.equ	ACSR,	0x08
.equ	DIDR0,	0x14		; Digital Input Disable Register 0
.equ	 ADC0D,	5
.equ	 ADC1D,	2
.equ	 ADC2D,	4
.equ	 ADC3D,	3
.equ	PCMSK,	0x15
.equ	PINB,	0x16
.equ	DDRB,	0x17
.equ	PORTB,	0x18
.equ	 LED_GREEN,	0	; PB0: green LED
.equ	 LED_RED,	1	; PB1: red LED
.equ	 CALIBRATE,	2	; PB2: calibration input
.equ	EECR,	0x1c		; EEPROM Control Register
.equ	 EEPM1,	5		; - Programming Mode Bit 1
.equ	 EEPM0,	4		; - Programming Mode Bit 0
.equ	 EERIE,	3		; - Ready Interrupt Enable
.equ	 EEMPE,	2		; - Master Program Enable
.equ	 EEPE,	1		; - Program Enable
.equ	 EERE,	0		; - Read Enable
.equ	EEDR,	0x1d		; EEPROM Data Register
.equ	EEARL,	0x1e		; EEPROM Address Register
.equ	WDTCR,	0x21
.equ	CLKPR,	0x26
.equ	GTCCR,	0x28
.equ	OCR0B,	0x29
.equ	DWDR,	0x2e
.equ	TCCR0A,	0x2f
.equ	OSCCAL,	0x31
.equ	TCNT0,	0x32		; Timer Counter 0
.equ	TCCR0B,	0x33		; Timer/Counter0 Control Register
.equ	MCUSR,	0x34
.equ	MCUCR,	0x35		; MCU Control Register
.equ	OCR0A,	0x36
.equ	SPMCSR,	0x37
.equ	TIFR0,	0x38
.equ	TIMSK0,	0x39		; Timer/Counter Interrupt Mask Register
.equ	GIFR,	0x3a
.equ	GIMSK,	0x3b		; General Interrupt Mask Register
.equ	SPL,	0x3d
.equ	SREG,	0x3f		; Status Register
.equ	RAMEND,	0x9f

; ------------------------------------------------------------
; Register definitions
; ------------------------------------------------------------
// multiplication
#define	A0		r0
#define	A1		r1
#define	B0		r2
#define	B1		r3
#define	PROD0		r4
#define	PROD1		r5
#define	PROD2		r6
#define	PROD3		r7

// division
#define	VALUE0		r0
#define	VALUE1		r1
#define	VALUE2		r2
#define	VALUE3		r3
#define	DIV0		r4
#define	DIV1		r5
#define	QUOT0		r6
#define	QUOT1		r7
#define	REM0		r30
#define	REM1		r31

// calibration data
#define	OFFSET2_L	r8
#define	OFFSET2_H	r9
#define	GAIN2_L		r10
#define	GAIN2_H		r11
#define	OFFSET3_L	r12
#define	OFFSET3_H	r13
#define	GAIN3_L		r14
#define	GAIN3_H		r15

// misc.
#define	TMP		r16	// temporary usage, not saved
#define	CHANNEL		r17	// 02 ==> ADC2/PB4,  03 ==> ADC3/PB3
#define	ADC_L		r18	// A/D result - low byte
#define	ADC_H		r19	// A/D result - high byte
#define	V2_L		r20	// charger voltage - low byte
#define	V2_H		r21	// charger voltage - high byte
#define	V3_L		r22	// battery voltage - low byte
#define	V3_H		r23	// battery voltage - high byte
#define	DELAY		r24	// units of 10msec
#define	LED		r25	// 00 ==> green LED, 01 ==> red LED
#define	COUNT		r26	// flash count

// EEPROM access
#define	EE_ADDR		r28	// EEPROM address
#define	EE_DATA		r29	// EEPROM data byte

; ------------------------------------------------------------
; Vector table
; ------------------------------------------------------------
	rjmp	reset		; RESET
	reti			; External Interrupt Request 0
	reti			; Pin Change Interrupt Request 0
	reti			; Timer/Counter Overflow
	reti			; EEPROM Ready
	reti			; Analog Comparator
	reti			; Timer/Counter Compare Match A
	reti			; Timer/Counter Compare Match B
	reti			; Watchdog Time-out
	reti			; ADC Conversion Complete

; ------------------------------------------------------------
; Program start
; ------------------------------------------------------------
reset:
	; initialize stack
	ldi	TMP, RAMEND
	out	SPL, TMP

	; control power consumption
	ldi	TMP, 0x20
	out	MCUCR, TMP	; enable idle sleep mode
	sbi	ACSR, 7		; switch off ACD

	; configure digital I/O pins
	sbi	DDRB, LED_GREEN	; digital output - green LED
	sbi	DDRB, LED_RED	; digital output - red LED
	sbi	PORTB, CALIBRATE; digital input  - calibration input

	; initializer timer
	ldi	TMP, 0x02
	out	TCCR0A, TMP	; CTC mode
	ldi	TMP, 0x05
	out	TCCR0B, TMP	; set prescaler to 1024
	ldi	TMP, ((CLOCK/TIMER)+512)/1024
	out	OCR0A, TMP	; <TIMER> interrupts per second
	ldi	TMP, 0x04
	out	TIMSK0, TMP	; enable timer interrupt

	; initialize A/D convertor
	sbi	DIDR0, ADC2D	; analog input - charger voltage
	sbi	DIDR0, ADC3D	; analog input - battery voltage
	ldi	TMP, 0x08+ADC_PRESCALER
	out	ADCSRA, TMP	; interrupt enabled, ADC disabled

	; enabled interrupts
	sei

; ------------------------------------------------------------
; Main loop.
; ------------------------------------------------------------
main:
	; blink once
	sbi	PORTB, LED_RED
	sleep
	sleep
	cbi	PORTB, LED_RED
	sleep

restart:
	; calculate OFFSET2/GAIN2
	ldi	EE_ADDR, 0
	rcall	cal_data
	movw	OFFSET2_L, ADC_L
	movw	GAIN2_L, QUOT0

	; calculate OFFSET3/GAIN3
	rcall	cal_data
	movw	OFFSET3_L, ADC_L
	movw	GAIN3_L, QUOT0

pause:
	ldi	DELAY, 200	; 2 sec pause
	rcall	delay10

loop:
	; V3 = voltage of channel 3 (battery) in units of 0.1/256 Volt
	ldi	CHANNEL, 3
	rcall	adc
	sub	ADC_L, OFFSET3_L
	sbc	ADC_H, OFFSET3_H
	movw	A0, ADC_L
	movw	B0, GAIN3_L
	rcall	mul16
	mov	V3_L, PROD1
	mov	V3_H, PROD2

	; V2 = voltage of channel 2 (charger) in units of 0.1/256 Volt
	ldi	CHANNEL, 2
	rcall	adc
	sub	ADC_L, OFFSET2_L
	sbc	ADC_H, OFFSET2_H
	movw	A0, ADC_L
	movw	B0, GAIN2_L
	rcall	mul16
	mov	V2_L, PROD1
	mov	V2_H, PROD2

	; show battery voltage with red LED (units: 0.1V)
	mov	COUNT, V3_H
	sbrc	V3_L, 7
	inc	COUNT		; rounding
	ldi	LED, 1		; red
	rcall	flash3

	; V2 = charge current in units of 10/256 mA
	sub	V2_L, V3_L
	sbc	V2_H, V3_H
	brcs	pause		; negative current

	; show charge current with green LED (units: 10mA)
	mov	COUNT, V2_H
	sbrc	V2_L, 7
	inc	COUNT		; rounding
	tst	COUNT
	breq	pause		; current < 10mA
	ldi	LED, 0		; green
	rcall	flash2

	rjmp	loop

cal_data:
	; read 8V & 12V samples from EEPROM
	rcall	eeprom_read	; 8V sample
	mov	ADC_L, EE_DATA
	rcall	eeprom_read
	mov	ADC_H, EE_DATA
	rcall	eeprom_read	; 12V sample
	mov	DIV0, EE_DATA
	rcall	eeprom_read
	mov	DIV1, EE_DATA

	; calculate offset: 3*v8 - 2*v12
	movw	QUOT0, ADC_L
	lsl	ADC_L
	rol	ADC_H
	add	ADC_L, QUOT0
	adc	ADC_H, QUOT1	; 3*v8
	sub	ADC_L, DIV0
	sbc	ADC_H, DIV1
	sub	ADC_L, DIV0
	sbc	ADC_H, DIV1	; 3*v8 - 2*v12

	; calculate gain: 120*65536 / (v12-offset)
	clr	VALUE0
	clr	VALUE1
	ldi	TMP, 120
	mov	VALUE2, TMP
	clr	VALUE3		; 120*65536
	sub	DIV0, ADC_L
	sbc	DIV1, ADC_H	; v12-offset
	rcall	div32		; 120*65536 / (v12-offset)
	ret

; ------------------------------------------------------------
; Calibration.
; ------------------------------------------------------------
calibrate:
	; clear stack
	ldi	TMP, RAMEND
	out	SPL, TMP

	; turn off LEDs
	cbi	PORTB, LED_GREEN
	cbi	PORTB, LED_RED

	; read both channels
	ldi	CHANNEL, 2
	rcall	adc
	movw	V2_L, ADC_L
	ldi	CHANNEL, 3
	rcall	adc
	movw	V3_L, ADC_L

	; differentiate between 8V and 12V
	sbrc	V2_H, 2		; test bit 10
	rjmp	v12
v8:
	sbi	PORTB, LED_GREEN
	ldi	EE_ADDR, 0
	rjmp	save
v12:
	sbi	PORTB, LED_RED
	ldi	EE_ADDR, 2
save:
	; write V2 & V3 to EEPROM:
	;   0: channel 2 - 8V
	;   2: channel 2 - 12V
	;   4: channel 3 - 8V
	;   6: channel 3 - 12V
	mov	EE_DATA, V2_L
	rcall	eeprom_write
	mov	EE_DATA, V2_H
	rcall	eeprom_write
	subi	EE_ADDR, -2
	mov	EE_DATA, V3_L
	rcall	eeprom_write
	mov	EE_DATA, V3_H
	rcall	eeprom_write

	; keep the LED turned on for 2 seconds
	ldi	DELAY, 200
0:	sleep
	dec	DELAY
	brne	0b
	cbi	PORTB, LED_GREEN
	cbi	PORTB, LED_RED
	rjmp	restart

; ------------------------------------------------------------
; Do one A/D conversion on channel CHANNEL and return the result in ADC.
; Use four times oversampling to get a resolution of 11 bit when there
; is some noise on the inputs.
; ------------------------------------------------------------
adc:
	sleep			; wait for the next timer interrupt
	out	ADMUX, CHANNEL	; Vcc reference, right adjust
	sbi	ADCSRA, 7	; enable ADC
	clr	ADC_L
	clr	ADC_H
0:	sleep			; start conversion
	in	TMP, ADCL	; get 10-bit result
	add	ADC_L, TMP
	in	TMP, ADCH
	ori	TMP, 0x40	; 4 samples
	adc	ADC_H, TMP
	brcc	0b
	asr	ADC_H		; drop one bit
	ror	ADC_L
	cbi	ADCSRA, 7	; disable ADC
	ret

; ------------------------------------------------------------
; Show three digits by flashing a LED.
; COUNT = value 0..255
; ------------------------------------------------------------
flash3:
	push	COUNT
	mov	TMP, COUNT
	ser	COUNT
0:	inc	COUNT
	subi	TMP, 100
	brcc	0b
	subi	TMP, -100
	rcall	flash1
	mov	COUNT, TMP
	rcall	flash2
	pop	COUNT
	ret

; ------------------------------------------------------------
; Show two digits by flashing a LED.
; COUNT = value 0..99
; ------------------------------------------------------------
flash2:
	push	COUNT
	mov	TMP, COUNT
	ser	COUNT
0:	inc	COUNT
	subi	TMP, 10
	brcc	0b
	subi	TMP, -10
	rcall	flash1
	mov	COUNT, TMP
	rcall	flash1
	pop	COUNT
	push	DELAY
	ldi	DELAY, 100		; pause between values: 1000 msec
	rcall	delay10
	pop	DELAY
	ret

; ------------------------------------------------------------
; Show one digit by flashing a LED. Zero is signalled by a long flash.
; COUNT = digit value 0..9
; ------------------------------------------------------------
flash1:
	push	COUNT
	push	DELAY
	tst	COUNT
	brne	0f
	inc	COUNT
	ldi	DELAY, 20		; long flash: 200 msec
	cpse	DELAY, DELAY
0:	ldi	DELAY, 3		; short flash: 30 msec
	sbrs	LED, 0
	sbi	PORTB, LED_GREEN
	sbrc	LED, 0
	sbi	PORTB, LED_RED
	rcall	delay10
	cbi	PORTB, LED_GREEN
	cbi	PORTB, LED_RED
	ldi	DELAY, 27		; pause between flashes: 270 msec
	rcall	delay10
	dec	COUNT
	brne	0b
	ldi	DELAY, 50		; pause between digits: 500 msec
	rcall	delay10
	pop	DELAY
	pop	COUNT
	ret

; ------------------------------------------------------------
; Delay DELAY times 10msec.
; ------------------------------------------------------------
delay10:
	push	DELAY
0:	sleep
	sbis	PINB, CALIBRATE
	rjmp	calibrate
	dec	DELAY
	brne	0b
	pop	DELAY
	ret

; ------------------------------------------------------------
; unsigned multiplication, 16x16 --> 32
; PROD = A * B
; ------------------------------------------------------------
mul16:
	push	COUNT
	clr	PROD3
	clr	PROD2
	movw	PROD0, B0
	ldi	COUNT, 16
	lsr	PROD1
	ror	PROD0
0:	brcc	1f
	add	PROD2, A0
	adc	PROD3, A1
1:	ror	PROD3
	ror	PROD2
	ror	PROD1
	ror	PROD0
	dec	COUNT
	brne	0b
	pop	COUNT
	ret

; ------------------------------------------------------------
; unsigned division, 32/15 --> 16+15
; QUOT = VALUE / DIV
; REM  = VALUE % DIV
; Note: the divisor must have bit 15 cleared!
; ------------------------------------------------------------
div32:
	push	COUNT
	push	VALUE2
	push	VALUE3
	clr	REM0
	clr	REM1
	ldi	COUNT, 32
	movw	QUOT0, VALUE0
	rol	QUOT0
	rol	QUOT1
0:	rol	VALUE2
	rol	VALUE3
	rol	REM0
	rol	REM1
	sub	REM0, DIV0
	sbc	REM1, DIV1
	brcc	1f
	add	REM0, DIV0
	adc	REM1, DIV1
1:	rol	QUOT0
	rol	QUOT1
	dec	COUNT
	brne	0b
	com	QUOT0
	com	QUOT1
	pop	VALUE3
	pop	VALUE2
	pop	COUNT
	ret

; ------------------------------------------------------------
; Read a byte from EEPROM address EE_ADDR into EE_DATA, and increment EE_ADDR
; ------------------------------------------------------------
eeprom_read:
	sbic	EECR, EEPE
	rjmp	eeprom_read	; wait for pending write
	out	EEARL, EE_ADDR
	sbi	EECR, EERE	; read byte into EEDR
	in	EE_DATA, EEDR
	inc	EE_ADDR
	ret

; ------------------------------------------------------------
; Write the byte EE_DATA to EEPROM address EE_ADDR, and increment EE_ADDR
; ------------------------------------------------------------
eeprom_write:
	sbic	EECR, EEPE
	rjmp	eeprom_write	; wait for pending write
	out	EEARL, EE_ADDR
	out	EEDR, EE_DATA
	in	TMP, SREG
	cli			; disable interrupts for the next 2 instructions
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	out	SREG, TMP
	inc	EE_ADDR
	ret

; ------------------------------------------------------------
; Identification
; ------------------------------------------------------------
	.align	4, 0xff
	.asciz	"Copyright 2005, Dick Streefland"
	.asciz	"2005-11-13"
	.align	4, 0xff
